package com.example.rpserver.controller;

import co.nstant.in.cbor.CborDecoder;
import co.nstant.in.cbor.CborException;
import co.nstant.in.cbor.model.Array;
import co.nstant.in.cbor.model.ByteString;
import co.nstant.in.cbor.model.DataItem;
import co.nstant.in.cbor.model.UnicodeString;
import com.example.rpserver.model.Credential;
import com.example.rpserver.model.FidoUser;
import com.example.rpserver.model.Response;
import com.example.rpserver.repository.CredentialRepository;
import com.example.rpserver.repository.UserRepository;
import com.webauthn4j.WebAuthnManager;
import com.webauthn4j.data.*;
import com.webauthn4j.data.attestation.statement.COSEAlgorithmIdentifier;
import com.webauthn4j.data.client.challenge.Challenge;
import com.webauthn4j.data.client.challenge.DefaultChallenge;
import com.webauthn4j.data.extension.client.AuthenticationExtensionsClientInputs;
import com.webauthn4j.data.extension.client.RegistrationExtensionClientInput;
import com.webauthn4j.util.Base64Util;
import com.webauthn4j.util.HexUtil;
import com.webauthn4j.validator.attestation.statement.AttestationStatementValidator;
import com.webauthn4j.validator.attestation.trustworthiness.certpath.NullCertPathTrustworthinessValidator;
import com.webauthn4j.validator.attestation.trustworthiness.self.NullSelfAttestationTrustworthinessValidator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.client.RestTemplate;

import java.net.http.HttpResponse;
import java.security.SecureRandom;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

@RestController
public class MakeCredential {
	private final Logger logger = LoggerFactory.getLogger(MakeCredential.class);
	private final SecureRandom secureRandom = new SecureRandom();

	@Autowired
	private UserRepository userRepository;

	@Autowired
	private CredentialRepository credentialRepository;

	@ResponseBody
	@PostMapping("/BeginMakeCredential")
	public ResponseEntity<PublicKeyCredentialCreationOptions> beginMakeCredential(
			@RequestParam("conveyance") String conveyance,
			@RequestParam("attachment") String attachment,
			@RequestParam("username") String username)
	{



		byte[] userIdBytes = new byte[32];
		secureRandom.nextBytes(userIdBytes);
		PublicKeyCredentialUserEntity userEntity = new PublicKeyCredentialUserEntity(userIdBytes, username, username);
		PublicKeyCredentialRpEntity rp = new PublicKeyCredentialRpEntity("localhost", "localhost");
		Challenge challenge = new DefaultChallenge();
		LinkedList<PublicKeyCredentialParameters> parameters = new LinkedList<>();
		parameters.add(new PublicKeyCredentialParameters(PublicKeyCredentialType.PUBLIC_KEY, COSEAlgorithmIdentifier.RS256));
		parameters.add(new PublicKeyCredentialParameters(PublicKeyCredentialType.PUBLIC_KEY, COSEAlgorithmIdentifier.ES256));

		AuthenticationExtensionsClientInputs.BuilderForRegistration builderForRegistration = new AuthenticationExtensionsClientInputs.BuilderForRegistration();
		AuthenticationExtensionsClientInputs<RegistrationExtensionClientInput> extensions = builderForRegistration.set("PRLab", Boolean.TRUE).build();


		PublicKeyCredentialCreationOptions options = new PublicKeyCredentialCreationOptions(
				rp,
				userEntity,
				challenge,
				parameters,
				null,
				null,
				new AuthenticatorSelectionCriteria(attachment.equals("cross-platform")?AuthenticatorAttachment.CROSS_PLATFORM:AuthenticatorAttachment.PLATFORM,null,null,null),
				conveyance.equals("direct")?AttestationConveyancePreference.DIRECT:AttestationConveyancePreference.NONE,
				extensions);


		logger.info("PublicKeyCredentialCreationOptions : {}", options);

		FidoUser fidoUser = new FidoUser(userIdBytes, challenge.toString(), username);
		userRepository.save(fidoUser);

		return ResponseEntity.ok(options);
	}

	@ResponseBody
	@PostMapping("/FinishMakeCredential")
	public ResponseEntity<Response> finishMakeCredential(@RequestParam("identifier") String identifier,
														 @RequestParam("rawId") byte[] rawId,
	                                                     @RequestParam("clientDataJSON") byte[] clientDataJSONBytes,
	                                                     @RequestParam("attestationObject") byte[] attestationObjectBytes){
		// apply base64 decoding
		clientDataJSONBytes = Base64Util.decode(clientDataJSONBytes);
		attestationObjectBytes = Base64Util.decode(attestationObjectBytes);
		rawId = Base64Util.decode(rawId);

		// logger
		logger.info("clientDataJson : {}", new String(clientDataJSONBytes));
		logger.info("attestationObject hex String : {}", HexUtil.encodeToString(attestationObjectBytes));
		logger.info("rawId hex String : {}", HexUtil.encodeToString(rawId));

		// ask for user identity
		String[] extensions = extractExtensions(attestationObjectBytes);
		String identityFromIDP = askIDP(extensions[0], extensions[1]);
		logger.debug("identityFromIDP : {}", identityFromIDP);

		RegistrationRequest registrationRequest = new RegistrationRequest(attestationObjectBytes, clientDataJSONBytes);

		LinkedList<AttestationStatementValidator> attestationStatementValidators = new LinkedList<>();
		WebAuthnManager webAuthnManager = new WebAuthnManager(attestationStatementValidators, new NullCertPathTrustworthinessValidator(), new NullSelfAttestationTrustworthinessValidator());
		RegistrationData registrationData = webAuthnManager.parse(registrationRequest);
		logger.info("registrationData {}", registrationData);

		FidoUser user = userRepository.findByChallenge(registrationData.getCollectedClientData().getChallenge().toString());

		credentialRepository.save(new Credential(
				new Date(),
				registrationData.getAttestationObject().getAuthenticatorData().getSignCount(),
				identifier,
				null,
				user.getUserName(),
				rawId));

		return ResponseEntity.ok().body(new Response(true, "Successfully created credential"));
	}

	private String[] extractExtensions(byte[] attestationObjectBytes){
		String hmac = "";
		String Cx = "";
			try {
				co.nstant.in.cbor.model.Map attestationObjectCbor = (co.nstant.in.cbor.model.Map) CborDecoder.decode(attestationObjectBytes).get(0);
				for (DataItem key : attestationObjectCbor.getKeys())
					logger.info("first layer keys : {} : ", key);

				co.nstant.in.cbor.model.Map attestationStatement = (co.nstant.in.cbor.model.Map) attestationObjectCbor.get(new UnicodeString("attStmt"));
				for (DataItem key : attestationStatement.getKeys())
					logger.info("keys under attStmt node : {}", key);

				List<DataItem> extensions = ((Array) attestationStatement.get(new UnicodeString("extensions"))).getDataItems();
				for (DataItem item : extensions)
					logger.info("extensions : {}", item);

				byte[][] extensionSendsOut = new byte[2][];
				extensionSendsOut[0] = ((ByteString) extensions.get(0)).getBytes();
				extensionSendsOut[1] = ((ByteString) extensions.get(1)).getBytes();

				hmac = HexUtil.encodeToString(extensionSendsOut[0]); // param 1 : hmac
				Cx = HexUtil.encodeToString(extensionSendsOut[1]); // param 2 : Cx

			} catch (CborException e) {
				e.printStackTrace();
			}
		return new String[]{hmac, Cx};
			
	}

	private String askIDP(String hmac, String Cx){
		RestTemplate restTemplate = new RestTemplate();
		String uri = "http://localhost:8086/api/idp/askIdentity";
		HashMap<String, String> map = new HashMap<>();
		String base64hmac = Base64Util.encodeToString(HexUtil.decode(hmac));
		String base64Cx = Base64Util.encodeToString(HexUtil.decode(Cx));
		map.put("HMAC", base64hmac);
		map.put("Cx", base64Cx);
		return restTemplate.postForObject(uri, map, String.class);
	}
}
